C     Copyright (C) 2004 B. Roukema et al. <shape-univ at astro.uni.torun.pl>
C     
C     This program is free software; you can redistribute it and/or modify
C     it under the terms of the GNU General Public License as published by
C     the Free Software Foundation; either version 2 of the License, or
C     (at your option) any later version.
C     
C     This program is distributed in the hope that it will be useful,
C     but WITHOUT ANY WARRANTY; without even the implied warranty of
C     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C     GNU General Public License for more details.
C     
C     You should have received a copy of the GNU General Public License
C     along with this program; if not, write to the Free Software
C     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
C  
C     See also http://www.gnu.org/licenses/gpl.html

C     --- calculate total number of "runs" for Poincare dodec case ----
      subroutine getnrun(dododec, dodran, resdodec,
     _     bmindodec,bmaxdodec,glmindodec,glmaxdodec,
     _     thmin,thmax,
     _     angmin,angmax,
     _     nangle, nrunin,
     _     nrun,nthdomax)
      
C     INPUTS:
      logical*4  dododec        ! Poincare dodecahedral
      logical*4  dodran         ! random: ignore most input parameters
      real*4     resdodec       ! resolution of grid in degrees
      real*4     bmindodec      ! minimum latitude (cf 90 = NGP)
      real*4     bmaxdodec,glmindodec,glmaxdodec ! etc
      real*4     thmin,thmax  ! min/max rotation of dodec
      real*4     angmin,angmax  ! min/max angular radius of circles
      integer*4  nangle         ! number of angular radii
      integer*4  nrunin         ! input suggestion of nrun (e.g. random)

C     OUTPUTS:
      integer*4  nrun
      integer*4  nthdomax ! same as nthdodec (in nruncomm)

C     INTERNAL:
      integer*4  ifirst
      parameter(pi=3.1415926536,pi_180=pi/180.0)

C     COMMON:
c      parameter(mxlb= 10000)    ! OK for resdodec = 1 deg
c     --- for single circle pairs - 27.02.2004 ---
      parameter(mxlb= 200000)    ! OK for resdodec = 1 deg
      real*4     gl_ilb(mxlb),gb_ilb(mxlb)
c      parameter(mxrun= 1000000)
      include 'npar.dec'
      parameter(mxrun=mxpar)

      integer*4  ilbirun(mxrun),ithirun(mxrun),iangirun(mxrun)
      real*4     delang ! angular radius interval

      common /nruncomm/ nlb,nthdodec,nang,delang, gl_ilb,gb_ilb,
     _     ilbirun,ithirun,iangirun


C     LOCAL:
      logical*4  test
      save ifirst
      data ifirst/0/

      if(.not.(dododec))then
         stop 'getnrun: dododec is .false.'
      endif

      if(ifirst.ne.12345)then
         ifirst=12345

         delb= resdodec
c         nb= max(1, int( (90.0-bmindodec)/delb ) )
         nb= max(1, int( (bmaxdodec-bmindodec)/delb ) )
         ii=0
c         nlequat= 360.0/resdodec

         do ib=1,nb
c            gb= 90.0- (real(ib)-0.5)*delb
            gb= bmaxdodec- (real(ib)-0.5)*delb

            dell= resdodec/cos(gb*pi_180)
c            nl= max(1, int( 360.0/dell ))
            nl= max(1, int( (glmaxdodec-glmindodec)/dell ))

            do il=1,nl
               ii=ii+1
               if(ii.gt.mxlb)then
                  print*,'getnrun ERROR: ii,mxlb=',ii,mxlb
               else
c                  gl_ilb(ii)= (real(il)-0.5)*dell
                  gl_ilb(ii)= glmindodec + (real(il)-0.5)*dell
                  gb_ilb(ii)= gb
               endif
            enddo
         enddo
         if(ii.gt.mxlb)then
            nlb=mxlb
            print*,'getnrun ERROR: ii,mxlb=',ii,mxlb            
         else
            nlb= ii
         endif

      endif

c      nthdodec= max(1, int( 72.0 / resdodec ))! 72.0 is hardwired for Poinc dodec

      nthdodec= max(1, int( (thmax-thmin) / resdodec )) !

      nthdomax= nthdodec
c      print*,'nthdodec=',nthdodec


c      nang= max(1, int( (angmax-angmin)/resdodec ))
      nang= nangle
      if(nang.le.1)then
         delang=0.0
      else
         delang=(angmax-angmin)/real(nang-1)
      endif

      nrun= nlb *nthdodec *nang

c     --- 0.1.35 ---
      if(dodran)then
         nrun= min(nrunin,mxrun)
c         nlb=int(nrun/(max(1,nthdodec)*max(1,nang)))
         nlb=int(nrun/(max(1,nthdodec)*max(1,nang))) + 1  ! 0.1.36
         nrun= nlb *nthdodec *nang ! 0.1.36
      endif

      if(min(nthdodec,nang).lt.1)then
         print*,'getnrun: WARNING: nthdodec or nang < 1:',
     _        nthdodec,nang
      endif

      irun=0

c     --- 0.1.35 shifted iang from outermost loop to innermost loop ---
      do ilb=1,nlb
         do ith=1,nthdodec
            do iang=1,nang
               irun=irun+1
               if(irun.le.nrun)then
                  ilbirun(irun)=ilb
                  ithirun(irun)=ith
                  iangirun(irun)=iang
               endif
            enddo
         enddo
      enddo
      

      print*,'getnrun: nlb,nthdodec,nang=',nlb,nthdodec,nang

      return
      end





C     --- calculate data for a  "run" for Poincare dodec case ----
      subroutine getirun(dododec, dodran, resdodec,
c     _     bmindodec,bmaxdodec,glmindodec,glmaxdodec,
     _     thmin,
     _     angmin,angmax,
     _     irun, 
     _     gl,gb,th,ang, ilb,ith,iang,isco_)
      
C     INPUTS:
      logical*4  dododec        ! Poincare dodecahedral
      logical*4  dodran         ! random: ignore most input parameters
      real*4     resdodec       ! resolution of grid in degrees
      real*4     bmindodec      ! minimum latitude (cf 90 = NGP)
      real*4     bmaxdodec,glmindodec,glmaxdodec ! etc
      real*4     thmin,thmax  ! min/max rotation of dodec
      real*4     angmin,angmax  ! min/max angular radius of circles
      integer*4  irun

C     OUTPUTS:
      real*4     gl,gb,th,ang   ! g long, lat, theta, angle for this irun
      integer*4  ilb,ith,iang
      logical*4  isco_        ! is this a control run?

C     INTERNAL:
      parameter(thsafe= 10.0)  ! great circle degrees safe margin for ran theta
      parameter(pi=3.1415926536,pi_180=pi/180.0)

C     SAVED:
      real*4 gl_old,gb_old,th_old
      save   gl_old,gb_old,th_old


C     COMMON:
c      parameter(mxlb= 10000) ! OK for resdodec = 1 deg
      parameter(mxlb= 200000)    ! OK for resdodec = 1 deg
      real*4     gl_ilb(mxlb),gb_ilb(mxlb)
c      parameter(mxrun= 1000000)
      include 'npar.dec'
      parameter(mxrun=mxpar)

      integer*4  ilbirun(mxrun),ithirun(mxrun),iangirun(mxrun)
      real*4     delang ! angular radius interval

      common /nruncomm/ nlb,nthdodec,nang, delang, gl_ilb,gb_ilb,
     _     ilbirun,ithirun,iangirun


      if(.not.(dododec))then
         stop 'getirun: dododec is .false.'
      endif

      tolang= 0.01 ! angles smaller than 1 degree are clearly meaningless


c      ilb= mod( irun-1, nlb ) +1
c      ithdodec= mod( irun - ilb -1, nlb*nthdodec ) +1
c      iang= mod( irun - ilb -(ithdodec-1)*nthdodec -1, 
c     _     nlb*nthdodec*nang ) +1
      

      ilb=ilbirun(irun)
      ith=ithirun(irun)
      iang= iangirun(irun)

      if(ilbirun(irun).lt.1)print*,'getirun: ERROR: ',
     _     'irun,ilbirun(irun)=',irun,ilbirun(irun)

      gl= gl_ilb(ilbirun(irun))
      gb= gb_ilb(ilbirun(irun))
c     th= (real(ithirun(irun))-0.5) *resdodec


c      ang= angmax - (real(iangirun(irun))-0.5)* delang  ! start from angmax
c     -- 04.03.2004 - should include endpoints see definition of delang --
      ang= angmax - (real(iangirun(irun)-1))* delang ! start from angmax


      th= thmin+ (real(ithirun(irun))-0.5) *resdodec

c     --- 19.07.2006 boud: circles-0.1.35:  .and.ilb.gt.1 ---
c     --- 29.06.2006 boud: circles-0.1.30 ---
      isco_ =.false.
      if(dodran.and.ilb.gt.1)then
         isco_ =.true.
         if(ith.eq.1.and.iang.eq.1)then ! first for this (gl,gb) pair
c     --- flush ran generator ---
            xx=rnd(0)*26
            do i=1,int(xx) 
               xx=rnd(0)
            enddo
            
            call spher2(gl,gb)
            
c         th= rnd(0)*360.0
c     --- 17.07.2006 boud: circles-0.1.35 ---
c     -     th is expressed in great circle degrees  (?) -
            if(ang.gt.tolang)then
               xx=rnd(0)*(
     _              max(180.0, 360.0 - thsafe/sin(ang*pi_180)))
               th=thmin + xx
               if(th.gt.360.0)th=th-360.0
            else
               th= rnd(0)*360.0
            endif

            gl_old=gl
            gb_old=gb
            th_old=th
         else   ! only randomise gl,gb,th;  not  ang
            gl=gl_old
            gb=gb_old
            th=th_old
         endif
      endif


      return
      end

